#include"dumberdowner.h"

#include<stdlib.h>
#include<assert.h>

// This is the dumbing down pass.
//
// Complex expressions are to be "broken down" into simpler ones until the AST
// can be trivially translated to the target architecture.
//
// This pass, along with CG is strictly dependent on x86 and will fail for
// any other architecture.

#include"x86.h"

static void varify_change_usedefs(AST *e, AST *oldStmt, AST *newStmt) {
	if(e->nodeKind == AST_EXPR_VAR) {
		if(e->exprVar.thing->kind != VARTABLEENTRY_VAR) {
			return;
		}
		
		UseDef *swapWithPrev = NULL;
		UseDef *swapWith = NULL;
		
		UseDef *udPrev = NULL;
		UseDef *ud = e->exprVar.thing->data.var.usedefFirst;
		while(ud && ud->use != e) {
			if(ud->stmt == oldStmt && !swapWith) {
				swapWithPrev = udPrev;
				swapWith = ud;
			}
			
			udPrev = ud;
			ud = ud->next;
		}
		
		if(ud) {
			assert(ud->stmt == oldStmt);
			ud->stmt = newStmt;
			
			// We must make sure all newStmt usedefs are before the oldStmt ones, so swap
			if(swapWith) {
				assert(!!swapWithPrev + !!udPrev != 0);
				
				if(swapWithPrev) {
					swapWithPrev->next = ud;
				} else {
					e->exprVar.thing->data.var.usedefFirst = ud;
				}
				
				if(udPrev) {
					udPrev->next = swapWith;
				} else {
					e->exprVar.thing->data.var.usedefFirst = swapWith;
				}
				
				UseDef *temp = ud->next;
				ud->next = swapWith->next;
				swapWith->next = temp;
				
				assert(!!ud->next + !!swapWith->next != 0);
				if(!swapWith->next) {
					e->exprVar.thing->data.var.usedefLast = swapWith;
				}
				if(!ud->next) {
					e->exprVar.thing->data.var.usedefLast = ud;
				}
			}
		}
	} else if(e->nodeKind == AST_EXPR_BINARY_OP) {
		varify_change_usedefs(e->exprBinOp.operands[0], oldStmt, newStmt);
		varify_change_usedefs(e->exprBinOp.operands[1], oldStmt, newStmt);
	} else if(e->nodeKind == AST_EXPR_UNARY_OP) {
		varify_change_usedefs(e->exprUnOp.operand, oldStmt, newStmt);
	} else if(e->nodeKind == AST_EXPR_CALL) {
		varify_change_usedefs(e->exprCall.what, oldStmt, newStmt);
		
		int argCount = e->exprCall.what->expression.type->function.argCount;
		for(int i = 0; i < argCount; i++) {
			varify_change_usedefs(e->exprCall.args[i], oldStmt, newStmt);
		}
	} else if(e->nodeKind == AST_EXPR_ARRAY) {
		int sz = e->exprArray.type->array.length;
		for(int i = 0; i < sz; i++) {
			varify_change_usedefs(e->exprArray.items[i], oldStmt, newStmt);
		}
	} else if(e->nodeKind == AST_EXPR_CAST) {
		varify_change_usedefs(e->exprCast.what, oldStmt, newStmt);
	}
}

/* Split away complex expression into a new local variable */
static AST *varify(AST *tlc, AST *chunk, AST *stmtPrev, AST *stmt, AST *e) {
	VarTableEntry *vte = malloc(sizeof(*vte));
	vte->kind = VARTABLEENTRY_VAR;
	vte->type = e->expression.type;
	vte->data.var.color = 0;
	vte->data.var.degree = 0;
	vte->data.var.priority = 0;
	vte->data.var.reachingDefs = NULL;
	
	// Add to var array
	tlc->chunk.vars = realloc(tlc->chunk.vars, sizeof(*tlc->chunk.vars) * (++tlc->chunk.varCount));
	tlc->chunk.vars[tlc->chunk.varCount - 1] = vte;
	
	// Alter AST
	
	ASTExprVar *ev[2];
	for(int i = 0; i < 2; i++) {
		ev[i] = malloc(sizeof(ASTExprVar));
		ev[i]->nodeKind = AST_EXPR_VAR;
		ev[i]->type = e->expression.type;
		ev[i]->thing = vte;
	}
	
	ASTStmtAssign *assign = malloc(sizeof(*assign));
	assign->nodeKind = AST_STMT_ASSIGN;
	assign->what = (AST*) ev[0];
	assign->to = e;
	
	if(stmtPrev) {
		stmtPrev->statement.next = (AST*) assign;
	} else {
		chunk->chunk.statementFirst = (AST*) assign;
	}
	assign->next = stmt;
	
	// Assemble ud-chain
	
	UseDef *ud[2];
	ud[0] = malloc(sizeof(UseDef));
	ud[1] = malloc(sizeof(UseDef));
	
	ud[0]->stmt = (AST*) assign;
	ud[0]->use = assign->what;
	ud[0]->def = (AST*) assign;
	ud[0]->next = ud[1];
	
	ud[1]->stmt = stmt;
	ud[1]->use = (AST*) ev[1];
	ud[1]->def = (AST*) assign;
	ud[1]->next = NULL;
	
	vte->data.var.usedefFirst = ud[0];
	vte->data.var.usedefLast = ud[1];
	
	// Correct the ud-chain of variables used in varified expression `e`
	
	varify_change_usedefs(e, stmt, (AST*) assign);
	
	return (AST*) ev[1];
}

static AST *xopify(AST *tlc, AST *chunk, AST *stmtPrev, AST *stmt, AST *e) {
	return varify(tlc, chunk, stmtPrev, stmt, e);
}

static UseDef *get_last_usedef_before_stmt(AST *tlc, VarTableEntry *v, AST *stmt) {
	assert(v->kind == VARTABLEENTRY_VAR);
	
	UseDef *ud = v->data.var.usedefFirst;
	
	if(!ud) {
		return NULL;
	}
	
	UseDef *udPrev = NULL;
	
	while(ud) {
		if(ast_stmt_is_after(tlc, ud->stmt, stmt)) {
			break;
		}
		udPrev = ud;
		ud = ud->next;
	}
	
	return udPrev;
}

static int dumben_chunk(AST *tlc, AST *chu) {
	AST *sPrev = NULL;
	AST *s = chu->chunk.statementFirst;
	
	int effective = 0;
	
	while(s) {
	
		if(s->nodeKind == AST_STMT_IF) {
			
			AST *e = s->stmtIf.expression;
			
			if(e->nodeKind == AST_EXPR_BINARY_OP && binop_is_comparison(e->exprBinOp.operator)) {
			
				if(is_xop(e->exprBinOp.operands[0]) == XOP_NOT_XOP) {
					e->exprBinOp.operands[0] = xopify(tlc, chu, sPrev, s, e->exprBinOp.operands[0]);
					effective = 1;
				}
				
				if(is_xop(e->exprBinOp.operands[1]) == XOP_NOT_XOP) {
					e->exprBinOp.operands[1] = xopify(tlc, chu, sPrev, s, e->exprBinOp.operands[1]);
					effective = 1;
				}
				
				if(is_xop(e->exprBinOp.operands[0]) == XOP_MEM && is_xop(e->exprBinOp.operands[1]) == XOP_MEM) {
					// Can't have two mems; put one in var
					
					e->exprBinOp.operands[1] = varify(tlc, chu, sPrev, s, e->exprBinOp.operands[1]);
					effective = 1;
				}
				
			} else {
				
				s->stmtIf.expression = varify(tlc, chu, sPrev, s, e);
				effective = 1;
				
			}
			
			effective |= dumben_chunk(tlc, s->stmtIf.then);
			
		} else if(s->nodeKind == AST_STMT_LOOP) {
			
			effective |= dumben_chunk(tlc, s->stmtLoop.body);
			
		} else if(s->nodeKind == AST_STMT_ASSIGN) {
			
			if(s->stmtAssign.what->nodeKind == AST_EXPR_UNARY_OP && s->stmtAssign.what->exprUnOp.operator == UNOP_DEREF
				&& s->stmtAssign.to->nodeKind == AST_EXPR_UNARY_OP && s->stmtAssign.to->exprUnOp.operator == UNOP_DEREF) {
				
				s->stmtAssign.to = varify(tlc, chu, sPrev, s, s->stmtAssign.to);
				effective = 1;
			}
			
			if(s->stmtAssign.what->nodeKind == AST_EXPR_VAR && s->stmtAssign.what->exprVar.thing->kind == VARTABLEENTRY_VAR && s->stmtAssign.to->nodeKind == AST_EXPR_UNARY_OP && s->stmtAssign.to->exprUnOp.operator == UNOP_NEGATE && (s->stmtAssign.to->exprUnOp.operand->nodeKind != AST_EXPR_VAR || s->stmtAssign.what->exprVar.thing != s->stmtAssign.to->exprUnOp.operand->exprVar.thing)) {
				// Turn this:
				//  a = -b
				// into
				//  a = b
				//  a = -a
				// While keeping UD-chain valid.
				
				// Currently only for variables. While this could work for any XOP, it requires the ability to deep-copy `a`, which is not trivial when taking into account UD-chains.
				// TODO: Scrap incremental UD-chain modifications. Simply rebuild them after each pass.
				
				ASTExprVar *ev[2];
				for(int i = 0; i < 2; i++) {
					ev[i] = malloc(sizeof(ASTExprVar));
					ev[i]->nodeKind = AST_EXPR_VAR;
					ev[i]->type = s->stmtAssign.what->expression.type;
					ev[i]->thing = s->stmtAssign.what->exprVar.thing;
				}
				
				AST *negation = s->stmtAssign.to;
				
				s->stmtAssign.to = negation->exprUnOp.operand;
				negation->exprUnOp.operand = ev[0];
				
				AST *assign2 = malloc(sizeof(ASTStmtAssign));
				assign2->nodeKind = AST_STMT_ASSIGN;
				assign2->stmtAssign.what = ev[1];
				assign2->stmtAssign.to = negation;
				
				assign2->statement.next = s->statement.next;
				s->statement.next = assign2;

				UseDef *link = get_last_usedef_before_stmt(tlc, ev[0]->thing, assign2->statement.next);
				
				UseDef *ud[2];
				ud[0] = malloc(sizeof(UseDef));
				ud[1] = malloc(sizeof(UseDef));

				ud[0]->stmt = assign2;
				ud[0]->use = (AST*) ev[1];
				ud[0]->def = s;
				ud[0]->next = ud[1];

				ud[1]->stmt = assign2;
				ud[1]->use = (AST*) ev[0];
				ud[1]->def = s;
				ud[1]->next = link->next;
				
				link->next = ud[0];
				
				effective = 1;
			}
			
		} else if(s->nodeKind == AST_STMT_EXPR && s->stmtExpr.expr->nodeKind == AST_EXPR_CALL) {
			
			int argCount = s->stmtExpr.expr->exprCall.what->expression.type->function.argCount;
			
			for(int i = 0; i < argCount; i++) {
				if(is_xop(s->stmtExpr.expr->exprCall.args[i]) == XOP_NOT_XOP) {
					s->stmtExpr.expr->exprCall.args[i] = xopify(tlc, chu, sPrev, s, s->stmtExpr.expr->exprCall.args[i]);
					effective = 1;
				}
			}
			
		}
		
		sPrev = s;
		s = s->statement.next;
	}
	
	return effective;
}

void dumben_go(AST* tlc) {
	while(dumben_chunk(tlc, tlc));
}
